@@ -13,6 +13,7 @@ from lxml import etree
 from lxml.etree import XSLTParseError
 from Crypto.Hash import SHA256
 
+from .contants import TRASLADO, EGRESO, INGRESO, PAGO, COMPLEMENTO_CARTA_PORTE, COMPLEMENTO_PAGO
 from .utils import replace
 
 import os
@@ -36,13 +37,15 @@ class AbstractGeneratorXml(object):
         self.dir_xml = os.path.dirname(os.path.abspath(__file__))
         self.dom = parse(self.dir_xml + '/comprobante.xml')
         self.algoritmoSha = None
-        self.tipo_objeto = self.data.get('tipo')
+        self.tipo_objeto = self.data.get('tipo') # I, E, P, T
+        self.complemento = self.data.get('complemento', '') # Pago, CartaPorte
+        self.tipo_transporte = self.data.get('tipo_transporte', '')
         self.version = ''
-        self.__cadena_original = ''
+        self.cadena_original_value = ''
 
     @property
-    def cadena_original(self):
-        return self.__cadena_original
+    def cadena_original(self) -> str:
+        return self.cadena_original_value
 
     @abstractmethod
     def generate_xml(self):
@@ -98,10 +101,10 @@ class AbstractGeneratorXml(object):
             certificado_file_pem_bytes = b64decode(certificado_file_pem64)
             key = RSA.importKey(certificado_file_pem_bytes)
             logging.info('OBTENIENDO LA CADENA ORIGINAL DEL XML')
-            self.__cadena_original = self.get_cadena_original(self.get_xml())
+            self.cadena_original_value = self.get_cadena_original(self.get_xml())
             logging.info('LA CADENA ORIGINAL DEL XML')
             digest = self.algoritmoSha.new()
-            digest.update(self.__cadena_original)
+            digest.update(self.cadena_original_value)
             signer = PKCS1_v1_5.new(key)
             sign = signer.sign(digest)
             sello = b64encode(sign)
@@ -116,37 +119,17 @@ class AbstractGeneratorXml(object):
                 xml = xml.encode('utf-8')
             xml = etree.fromstring(xml)
             logging.info('PARSE DE LA CADENA ORIGINAL DEL XML')
-            url = 'http://www.sat.gob.mx/sitio_internet/cfd/3/cadenaoriginal_3_3/cadenaoriginal_3_3.xslt'
-            res = None
-            try:
-                res = requests.get(url, timeout=240, stream=True)
-            except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
-                logging.error(u"Hubo un error en el servidor del SAT, por favor intente mas tarde: " + str(e))
-                raise Exception(u"Hubo un error en el servidor del SAT, por favor intente mas tarde: " + str(e))
-            except Exception as e:
-                logging.error(u"Hubo un error en el servidor del SAT, por favor intente mas tarde: " + str(e))
-                raise Exception(u"Hubo un error en el servidor del SAT, por favor intente mas tarde: " + str(e))
-            if res is None:
-                logging.error(u"Hubo un error en el servidor del SAT, por favor intente mas tarde.")
-                raise Exception(u"Hubo un error en el servidor del SAT, por favor intente mas tarde.")
-            elif res.content is None:
-                logging.error(u"Hubo un error en el servidor del SAT, por favor intente mas tarde.")
-                raise Exception(u"Hubo un error en el servidor del SAT, por favor intente mas tarde.")
-            elif res.content is b'' or len(res.content) == 0:
-                logging.error(u"Hubo un error en el servidor del SAT, por favor intente mas tarde.")
-                raise Exception(u"Hubo un error en el servidor del SAT, por favor intente mas tarde.")
-            xslt = etree.fromstring(res.content)
+            logging.info(self.dir_xml)
+            xslt = etree.parse(self.dir_xml + '/cadenaoriginal_3_3_local/cadenaoriginal_3_3.xslt')
             logging.info('XSLT CADENA ORIGINAL DEL XML')
             parser = etree.XSLT(xslt)
             logging.info('XSLT CADENA ORIGINAL DEL XML 2')
-            self.__cadena_original = parser(xml)
+            cadena_original = parser(xml)
             logging.info('XSLT CADENA ORIGINAL DEL XML 3')
-            return self.__cadena_original
-            # else:
-            #     raise ('Error with cadenaoriginal_3_3.xslt')
+            return cadena_original
         except XSLTParseError as e:
-            logging.error(u"Hubo un error en el servidor del SAT, por favor intente mas tarde: " + str(e))
-            raise Exception(u"Hubo un error en el servidor del SAT, por favor intente mas tarde: " + str(e))
+            logging.error(u"Hubo un error al generar la cadena original, por favor intente mas tarde: " + str(e))
+            raise Exception(u"Hubo un error al generar la cadena original, por favor intente mas tarde: " + str(e))
         except Exception as e:
             logging.error('Error al leer el archivo: cadenaoriginal_' + self.data.get('version') + '.xslt:' + str(e))
             raise Exception(e)
@@ -216,13 +199,26 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
     def generate_xml(self):
         self.reset_dom()
         dom = self.dom.childNodes[ 0 ]
+        # xmlns: cfdi = "http://www.sat.gob.mx/cfd/3"
+        # xmlns: xsi = "http://www.w3.org/2001/XMLSchema-instance"
+        # xmlns: cartaporte = "http://www.sat.gob.mx/CartaPorte"
+        # xsi: schemaLocation = "http://www.sat.gob.mx/cfd/3 http://www.sat.gob.mx/sitio_internet/cfd/3/cfdv33.xsd http://www.sat.gob.mx/CartaPorte http://www.sat.gob.mx/sitio_internet/cfd/CartaPorte/CartaPorte.xsd"
         if self.data.get('tipo_comprobante') == 4: #recibo donativo
             dom.setAttribute('xsi:schemaLocation',
                              'http://www.sat.gob.mx/cfd/3 ' +
                              'http://www.sat.gob.mx/sitio_internet/cfd/3/cfdv33.xsd ' +
                              'http://www.sat.gob.mx/donat ' +
                              'http://www.sat.gob.mx/sitio_internet/cfd/donat/donat11.xsd')
-            dom.setAttribute('xmlns:donat','http://www.sat.gob.mx/donat')
+            dom.setAttribute('xmlns:donat', 'http://www.sat.gob.mx/donat')
+        elif self.tipo_objeto == TRASLADO:
+            dom.setAttribute('xmlns:cfdi', 'http://www.sat.gob.mx/cfd/3')
+            dom.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
+            dom.setAttribute('xmlns:cartaporte', 'http://www.sat.gob.mx/CartaPorte')
+            dom.setAttribute('xsi:schemaLocation',
+                             'http://www.sat.gob.mx/cfd/3 ' +
+                             'http://www.sat.gob.mx/sitio_internet/cfd/3/cfdv33.xsd ' +
+                             'http://www.sat.gob.mx/CartaPorte ' +
+                             'http://www.sat.gob.mx/sitio_internet/cfd/CartaPorte/CartaPorte.xsd')
         else:
             dom.setAttribute('xsi:schemaLocation',
                              'http://www.sat.gob.mx/cfd/3 http://www.sat.gob.mx/sitio_internet/cfd/3/cfdv33.xsd http://www.sat.gob.mx/TimbreFiscalDigital http://www.sat.gob.mx/sitio_internet/cfd/timbrefiscaldigital/TimbreFiscalDigitalv11.xsd')
@@ -241,13 +237,20 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
         if self.data.get('confirmacion'):
             if len(self.data.get('confirmacion', [])) == 5:
                 dom.setAttribute('Confirmacion', self.data.get('confirmacion'))  # opcional
-        if self.tipo_objeto == 'E':
+        dom.appendChild(self.get_emisor())
+        dom.appendChild(self.get_receptor())
+
+        if self.tipo_objeto == EGRESO:
+            dom.appendChild(self.get_cfdi_relacionados())
+        if self.tipo_objeto == INGRESO and self.complemento == COMPLEMENTO_CARTA_PORTE and self.data.get('tipo_relacion') == '05':
+            # Agregar lo siguiente
+            # 1.- Información en el campo TipoEstación de la sección Ubicación con el valor “02” (Intermedia), contenida en el catálogo del complemento Carta Porte, c_TipoEstacion, publicado en el portal del SAT.
+            # 2.- Información en la sección TransporteFerroviario de la sección Mercancias.
             dom.appendChild(self.get_cfdi_relacionados())
         elif len(self.data.get('cfdi_relacionados', [])) > 0:
             dom.appendChild(self.get_cfdi_relacionados())
-        dom.appendChild(self.get_emisor())
-        dom.appendChild(self.get_receptor())
-        if (self.tipo_objeto == 'I') or (self.tipo_objeto == 'E' and not self.global_data.get('concepto_fijo')):
+
+        if (self.tipo_objeto == INGRESO) or (self.tipo_objeto == EGRESO and not self.global_data.get('concepto_fijo')):
             dom.setAttribute('FormaPago', self.data.get('forma_pago'))
             dom.setAttribute('CondicionesDePago', self.data.get('condicion_pago'))
             # dom.setAttribute('Descuento', '0.00') # 'Opcional'
@@ -266,12 +269,23 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
                 dom.appendChild(conceptos)
             if self.data.get('iva') != '0' and self.data.get('iva') != 0 and self.data.get('tipo_comprobante') != 3:
                 dom.appendChild(self.get_impuestos())
-        elif self.tipo_objeto == 'P':
+        elif self.tipo_objeto == PAGO:
             dom.setAttribute('SubTotal','0')
             dom.setAttribute('Moneda','XXX')
             dom.setAttribute('Total','0')
             dom.appendChild(self.get_concepto_fijo())
-        elif self.tipo_objeto == 'E' and self.global_data.get('concepto_fijo'):
+        elif self.tipo_objeto == TRASLADO:
+            dom.setAttribute('SubTotal','0')
+            dom.setAttribute('Moneda','XXX')
+            dom.setAttribute('Total','0')
+            dom.setAttribute('FormaPago', self.data.get('forma_pago'))
+            conceptos_list = self.data.get('conceptos', [])
+            if len(conceptos_list) > 0:
+                conceptos = self.dom.createElement('cfdi:Conceptos')
+                for concepto in conceptos_list:
+                    conceptos.appendChild(self.get_conceptos(concepto))
+                dom.appendChild(conceptos)
+        elif self.tipo_objeto == EGRESO and self.global_data.get('concepto_fijo'):
             dom.setAttribute('FormaPago',self.data.get('forma_pago'))
             dom.setAttribute('CondicionesDePago', self.data.get('condicion_pago'))
             dom.setAttribute('MetodoPago', self.data.get('metodo_pago'))
@@ -282,18 +296,19 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
             dom.setAttribute('Total',str(self.data.get('total')))
             dom.appendChild(self.get_concepto_fijo())
             dom.appendChild(self.get_impuestos())
-        dom.appendChild(self.get_complementos())
+        if self.data.get('tipo_comprobante') in [4, 5] or self.complemento in [COMPLEMENTO_PAGO, COMPLEMENTO_CARTA_PORTE]:
+            dom.appendChild(self.get_complementos())
         self.dom = dom
         return self.dom
 
     def get_cfdi_relacionados(self):
         try:
             relacionados = self.dom.createElement('cfdi:CfdiRelacionados')
-            if self.tipo_objeto == 'I':
+            if self.tipo_objeto == INGRESO:
                 relacionados.setAttribute('TipoRelacion', self.data.get('tipo_relacion'))
-            elif self.tipo_objeto == 'P':
+            elif self.tipo_objeto == PAGO:
                 relacionados.setAttribute('TipoRelacion','04')  # Sustitución de CFDI previos
-            elif self.tipo_objeto == 'E':
+            elif self.tipo_objeto == EGRESO:
                 relacionados.setAttribute('TipoRelacion', self.data.get('tipo_relacion'))
             cfdirelacionados = self.data.get('cfdi_relacionados')
             for uuid_cfdi in cfdirelacionados:
@@ -322,6 +337,8 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
             if self.data.get('receptor_residencia_fiscal') != 'MEX':
                 receptor.setAttribute('ResidenciaFiscal',self.data.get('receptor_residencia_fiscal'))
                 receptor.setAttribute('NumRegIdTrib',self.data.get('receptor_num_reg_id_trib'))
+            if self.tipo_objeto == TRASLADO and self.data.get('uso_cfdi') != 'P01':
+                raise Exception('El campo Uso CFDI debe contener el valor P01 - Por Definir')
             receptor.setAttribute('UsoCFDI', self.data.get('uso_cfdi'))
             return receptor
         except Exception as e:
@@ -334,11 +351,11 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
             concepto.setAttribute('ClaveProdServ','84111506')
             concepto.setAttribute('Cantidad','1')
             concepto.setAttribute('ClaveUnidad','ACT')
-            if self.tipo_objeto == 'P':
+            if self.tipo_objeto == PAGO:
                 concepto.setAttribute('Descripcion','Pago')
                 concepto.setAttribute('ValorUnitario','0')
                 concepto.setAttribute('Importe','0')
-            elif self.tipo_objeto == 'E':
+            elif self.tipo_objeto == EGRESO:
                 if self.data.get('descripcion', '') != '':
                     concepto.setAttribute('Descripcion',replace(self.data.get('descripcion', '').strip()))
                 else:
@@ -395,6 +412,8 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
                 complemento.appendChild(self.get_donatarias())
             elif self.data.get('tipo_comprobante') == 5:
                 complemento.appendChild(self.get_pagos())
+            elif self.complemento == COMPLEMENTO_CARTA_PORTE:
+                complemento.appendChild(self.get_complemento_cartaporte())
             return complemento
         except Exception as e:
             raise Exception("Ocurrio un error. Error::: " + str(e))
@@ -491,7 +510,8 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
                 raise Exception('Se debe especificar una descripción para poder timbrar y posteriormente crear una poliza.')
             cfdi_concepto.setAttribute('ValorUnitario', str(concepto.get('valor_unitario')))
             cfdi_concepto.setAttribute('Importe', str(concepto.get('importe')))
-            cfdi_concepto.appendChild(self.get_impuestos_concepto(concepto))
+            if self.tipo_objeto != TRASLADO and self.complemento != COMPLEMENTO_CARTA_PORTE:
+                cfdi_concepto.appendChild(self.get_impuestos_concepto(concepto))
             if concepto.get('informacion_aduanera'):
                 cfdi_concepto.appendChild(self.get_informacion_aduanera(concepto.get('informacion_aduanera')))
             if concepto.get('cuenta_predial'):
@@ -557,3 +577,122 @@ class GeneratorXmlThreeDotThree(AbstractGeneratorXml):
             return parte
         except Exception as e:
             raise Exception("Ocurrio un error. Error::: " + str(e))
+
+    def get_complemento_cartaporte(self):
+        try:
+            cartaporte = self.dom.createElement('cartaporte:CartaPorte')
+            cartaporte.setAttribute('Version', '1.0')
+            cartaporte.setAttribute('TranspInternac', self.data.get('TranspInternac'))
+            if self.data.get('TranspInternac') == 'Si':
+                if self.data.get('EntradaSalidaMerc', '') not in ['Entrada', 'Salida']:
+                    raise Exception('El valor del campo EntradaSalidaMerc debe ser: Entrada ó Salida')
+                cartaporte.setAttribute('EntradaSalidaMerc', self.data.get('EntradaSalidaMerc'))
+                cartaporte.setAttribute('ViaEntradaSalida', self.data.get('ViaEntradaSalida'))
+
+            cartaporte_ubicaciones = self.dom.createElement('cartaporte:Ubicaciones')
+            ubicaciones = self.data.get('ubicaciones', [])
+            total_distancia_recorrida = 0
+            for ubicacion_data in ubicaciones:
+                cartaporte_ubicacion = self.dom.createElement('cartaporte:Ubicacion')
+                if ubicacion_data.get('type') == 'origen':
+                    cartaporte_ubicacion_origen = self.dom.createElement('cartaporte:Origen')
+                    cartaporte_ubicacion_origen.setAttribute('RFCRemitente', ubicacion_data.get('RFCRemitente'))
+                    cartaporte_ubicacion_origen.setAttribute('NombreRemitente', ubicacion_data.get('NombreRemitente'))
+                    cartaporte_ubicacion_origen.setAttribute('FechaHoraSalida', ubicacion_data.get('FechaHoraSalida'))
+                    cartaporte_ubicacion.appendChild(cartaporte_ubicacion_origen)
+                elif ubicacion_data.get('type') == 'destino':
+                    cartaporte_ubicacion.setAttribute('DistanciaRecorrida', ubicacion_data.get('DistanciaRecorrida'))
+                    cartaporte_ubicacion_destino = self.dom.createElement('cartaporte:Destino')
+                    cartaporte_ubicacion_destino.setAttribute('RFCDestinatario', ubicacion_data.get('RFCDestinatario'))
+                    cartaporte_ubicacion_destino.setAttribute('NombreDestinatario', ubicacion_data.get('NombreDestinatario'))
+                    cartaporte_ubicacion_destino.setAttribute('FechaHoraProgLlegada', ubicacion_data.get('FechaHoraProgLlegada'))
+                    cartaporte_ubicacion.appendChild(cartaporte_ubicacion_destino)
+                    total_distancia_recorrida += int(ubicacion_data.get('DistanciaRecorrida'))
+
+                cartaporte_ubicacion_domicilio = self.dom.createElement('cartaporte:Domicilio')
+                cartaporte_ubicacion_domicilio.setAttribute('Calle', ubicacion_data.get('Calle'))
+                cartaporte_ubicacion_domicilio.setAttribute('NumeroExterior', ubicacion_data.get('NumeroExterior'))
+                cartaporte_ubicacion_domicilio.setAttribute('Colonia', ubicacion_data.get('Colonia'))
+                cartaporte_ubicacion_domicilio.setAttribute('Localidad', ubicacion_data.get('Localidad'))
+                cartaporte_ubicacion_domicilio.setAttribute('Referencia', ubicacion_data.get('Referencia'))
+                cartaporte_ubicacion_domicilio.setAttribute('Municipio', ubicacion_data.get('Municipio'))
+                cartaporte_ubicacion_domicilio.setAttribute('Estado', ubicacion_data.get('Estado'))
+                cartaporte_ubicacion_domicilio.setAttribute('Pais', ubicacion_data.get('Pais'))
+                cartaporte_ubicacion_domicilio.setAttribute('CodigoPostal', ubicacion_data.get('CodigoPostal'))
+                cartaporte_ubicacion.appendChild(cartaporte_ubicacion_domicilio)
+                cartaporte_ubicaciones.appendChild(cartaporte_ubicacion)
+            cartaporte.appendChild(cartaporte_ubicaciones)
+            if self.tipo_transporte in ['AutotransporteFederal', 'TransporteFerroviario']:
+                cartaporte.setAttribute('TotalDistRec', str(total_distancia_recorrida))
+
+            cartaporte_mercancias = self.dom.createElement('cartaporte:Mercancias')
+            mercancias = self.data.get('mercancias', [])
+            num_total_mercancias = 0
+            for mercancia_data in mercancias:
+                cartaporte_mercancia = self.dom.createElement('cartaporte:Mercancia')
+                cartaporte_mercancia.setAttribute('PesoEnKg', mercancia_data.get('PesoEnKg'))
+                cartaporte_mercancia.setAttribute('ValorMercancia', mercancia_data.get('ValorMercancia'))
+                cartaporte_mercancia.setAttribute('Moneda', mercancia_data.get('Moneda'))
+                cartaporte_mercancias.appendChild(cartaporte_mercancia)
+                num_total_mercancias += 1
+
+            cartaporte_mercancias.setAttribute('NumTotalMercancias', str(num_total_mercancias))
+            cartaporte.appendChild(cartaporte_mercancias)
+
+            if self.tipo_transporte not in ['AutotransporteFederal', 'TransporteMaritimo', 'TransporteAereo', 'TransporteFerroviario']:
+                raise Exception('Se debe registrar un tipo de transporte')
+            transporte_data = self.data.get('transporte')
+            if self.tipo_transporte == 'AutotransporteFederal':
+                cartaporte_tipo_transporte = self.dom.createElement('cartaporte:AutotransporteFederal')
+                cartaporte_tipo_transporte.setAttribute('PermSCT', transporte_data.get('PermSCT'))
+                cartaporte_tipo_transporte.setAttribute('NumPermisoSCT', transporte_data.get('NumPermisoSCT'))
+                cartaporte_tipo_transporte.setAttribute('NombreAseg', transporte_data.get('NombreAseg'))
+                cartaporte_tipo_transporte.setAttribute('NumPolizaSeguro', transporte_data.get('NumPolizaSeguro'))
+
+                cartaporte_identificacion_vehicular = self.dom.createElement('cartaporte:IdentificacionVehicular')
+                cartaporte_identificacion_vehicular.setAttribute('ConfigVehicular', transporte_data.get('ConfigVehicular'))
+                cartaporte_identificacion_vehicular.setAttribute('PlacaVM', transporte_data.get('PlacaVM'))
+                cartaporte_identificacion_vehicular.setAttribute('AnioModeloVM', transporte_data.get('AnioModeloVM'))
+                cartaporte_tipo_transporte.appendChild(cartaporte_identificacion_vehicular)
+
+                remolques = transporte_data.get('remolques', [])
+                cartaporte_remolques = self.dom.createElement('cartaporte:Remolques')
+                for remolque_data in remolques:
+                    cartaporte_remolque = self.dom.createElement('cartaporte:Remolque')
+                    cartaporte_remolque.setAttribute('SubTipoRem', remolque_data.get('SubTipoRem'))
+                    cartaporte_remolque.setAttribute('Placa', remolque_data.get('Placa'))
+                    cartaporte_remolques.appendChild(cartaporte_remolque)
+                cartaporte_tipo_transporte.appendChild(cartaporte_remolques)
+
+                cartaporte_mercancias.appendChild(cartaporte_tipo_transporte)
+
+            cartaporte_figura_transporte = self.dom.createElement('cartaporte:FiguraTransporte')
+            cartaporte_figura_transporte.setAttribute('CveTransporte', transporte_data.get('CveTransporte'))
+
+            operadores = transporte_data.get('operadores', [])
+            cartaporte_operadores = self.dom.createElement('cartaporte:Operadores')
+            for operador_data in operadores:
+                cartaporte_operador = self.dom.createElement('cartaporte:Operador')
+                cartaporte_operador.setAttribute('RFCOperador', operador_data.get('RFCOperador'))
+                cartaporte_operador.setAttribute('NumLicencia', operador_data.get('NumLicencia'))
+                cartaporte_operador.setAttribute('NombreOperador', operador_data.get('NombreOperador'))
+
+                cartaporte_operador_domicilio = self.dom.createElement('cartaporte:Domicilio')
+                cartaporte_operador_domicilio.setAttribute('Calle', operador_data.get('Calle'))
+                cartaporte_operador_domicilio.setAttribute('NumeroExterior', operador_data.get('NumeroExterior'))
+                cartaporte_operador_domicilio.setAttribute('Colonia', operador_data.get('Colonia'))
+                cartaporte_operador_domicilio.setAttribute('Localidad', operador_data.get('Localidad'))
+                cartaporte_operador_domicilio.setAttribute('Referencia', operador_data.get('Referencia'))
+                cartaporte_operador_domicilio.setAttribute('Municipio', operador_data.get('Municipio'))
+                cartaporte_operador_domicilio.setAttribute('Estado', operador_data.get('Estado'))
+                cartaporte_operador_domicilio.setAttribute('Pais', operador_data.get('Pais'))
+                cartaporte_operador_domicilio.setAttribute('CodigoPostal', operador_data.get('CodigoPostal'))
+                cartaporte_operador.appendChild(cartaporte_operador_domicilio)
+
+                cartaporte_operadores.appendChild(cartaporte_operador)
+            cartaporte_figura_transporte.appendChild(cartaporte_operadores)
+            cartaporte.appendChild(cartaporte_figura_transporte)
+
+            return cartaporte
+        except Exception as e:
+            raise Exception("Ocurrio un error al generar el complemento cartaporte. Error::: " + str(e))
